{-# LANGUAGE RecordWildCards #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE TypeSynonymInstances #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE RankNTypes #-}

-- | Hamiltonian path / HOX example.
--
-- TODO we probably should clean up and inline a bunch of stuff

module ADP.Fusion.Examples.HamiltonianPath where

import           Control.Monad.Primitive
import           Control.Monad.ST
import           Data.Array.Repa.Index
import           Data.Array.Repa.Shape
import           Data.Bits
import           Data.List (sortBy,tails,delete,minimumBy,(\\),foldl',maximumBy)
import           Data.Ord (comparing)
import           Data.Strict.Tuple hiding (snd)
import           Data.Vector.Fusion.Stream.Size
import           Debug.Trace
import qualified Data.Vector.Fusion.Stream.Monadic as M
import qualified Data.Vector.Fusion.Stream as S
import           Text.Printf
import           Data.Vector.Fusion.Util
import qualified Data.Vector as V
import qualified Data.Vector.Unboxed as VU
import           Data.ByteString (ByteString)
import           GHC.Exts
import qualified Numeric.Log as Log
import           Numeric.Log (Log)

import           Data.PrimitiveArray as PA hiding (map)
import qualified Data.PrimitiveArray.Zero as PAZ
import           Data.Array.Repa.ExtShape
import           ADP.Fusion

import           Data.Array.Repa.Index.BitSet (activeBits)



-- * Hamiltonian path index.

-- | hpiVis shall contain hpiHead and hpiLast

data HPI = HPI { hpiVis :: !Int, hpiHead :: !Int, hpiLast :: !Int }
         deriving (Eq)

instance Show HPI where
    show (HPI v h l) = printf "HPI {nodes: %s first: %d last: %d}" (show $ activeBits v) h l

--test = let x = HPI 63 0 0 in map (\z -> (z, toIndex x z)) $ S.toList $ rangeStream (HPI 0 0 0) x

instance Shape HPI where
    rank _ = 1
    zeroDim = HPI 0 0 0   -- having visited nothing
    unitDim = HPI 1 0 0   -- visit just node 0
    intersectDim = error "hpi/intersectDim"
    addDim (HPI v1 h1 l1) (HPI v2 h2 l2) = HPI (v1 `shiftL` popCount v2) (h1+h2) (l1+l2) -- TODO ???
    -- TODO addDIm is probably borked for anything non-trivial
    size (HPI v h l) = let m = popCount v + 1 in v * m * m
    sizeIsValid (HPI v h l) = v < (maxBound `div` (popCount v * popCount v))
    toIndex (HPI mv mh ml) (HPI v h l) = v * (popCount mv * popCount mv) + h * popCount mv + l
    fromIndex = error "hpi/fromIndex"
    inShapeRange = error "hpi/inShapeRange"
    listOfShape = error "hpi/listOfShape"
    shapeOfList = error "hpi/shapeOfList"
    deepSeq (HPI v h l) n = v `seq` h `seq` l `seq` n
    {-# INLINE [1] rank #-}
    {-# INLINE [1] zeroDim #-}
    {-# INLINE [1] unitDim #-}
    {-# INLINE [1] intersectDim #-}
    {-# INLINE [1] addDim #-}
    {-# INLINE [1] size #-}
    {-# INLINE [1] sizeIsValid #-}
    {-# INLINE [1] toIndex #-}
    {-# INLINE [1] fromIndex #-}
    {-# INLINE [1] inShapeRange #-}
    {-# NOINLINE listOfShape #-}
    {-# NOINLINE shapeOfList #-}
    {-# INLINE     deepSeq #-}

instance ExtShape HPI where
    subDim (HPI v1 h1 l1) (HPI v2 h2 l2) = HPI (v1 `shiftR` popCount v2) (h1-h2) (l1-l2)
    -- TODO subDim is probably borked for anything non-trivial
    rangeList (HPI 0 0 0) (HPI v _ _) =
      [ HPI v' h' l'
      | v' <- sortBy (comparing popCount) [0..v]
      , h' <- [0 .. popCount v -1]
      , l' <- [0 .. popCount v -1]
      ]
      {-
      HPI 0 0 0 :
      [ HPI v' h' l'
      | v' <- sortBy (comparing popCount) [0..v]
      , let as = activeBits v'
      , h' <- as
      , l' <- if length as == 1 then as else delete h' as
      ]
      -}
    rangeStream (HPI 0 0 0) (HPI v _ _) = M.fromList $
      -- the initial node
      [ HPI (0 `setBit` h') h' h'
      | h' <- activeBits v
      ] ++
      [ HPI v' h' l'
      | v' <- sortBy (comparing popCount) [0..v]
      , let as = activeBits v'
      , h' <- as
      , l' <- delete h' as
      ] ++ [HPI 0 0 0]
    {-# INLINE [1] subDim #-}
    {-# INLINE     rangeList #-}
    {-# INLINE     rangeStream #-}

data HPIo = HPIo { hpioVis :: !Int, hpioHead :: !Int, hpioLast :: !Int }
         deriving (Eq)

instance Show HPIo where
    show (HPIo v h l) = printf "HPIo {nodes: %s first: %d last: %d}" (show $ activeBits v) h l

instance Shape HPIo where
    rank _ = 1
    zeroDim = HPIo 0 (-1) (-1) -- having visited nothing
    unitDim = HPIo 1 0 0       -- visit just node 0
    intersectDim = error "hpio/intersectDim"
    addDim (HPIo v1 h1 l1) (HPIo v2 h2 l2) = HPIo (v1 `shiftL` popCount v2) (h1+h2) (l1+l2) -- TODO ???
    -- TODO addDIm is probably borked for anything non-trivial
    size (HPIo v h l) = v * popCount v * popCount v
    sizeIsValid (HPIo v h l) = v < (maxBound `div` (popCount v * popCount v))
    toIndex (HPIo mv mh ml) (HPIo v h l) = v * (popCount mv * popCount mv) + h * popCount mv + l
    fromIndex = error "hpio/fromIndex"
    inShapeRange = error "hpio/inShapeRange"
    listOfShape = error "hpio/listOfShape"
    shapeOfList = error "hpio/shapeOfList"
    deepSeq (HPIo v h l) n = v `seq` h `seq` l `seq` n
    {-# INLINE [1] rank #-}
    {-# INLINE [1] zeroDim #-}
    {-# INLINE [1] unitDim #-}
    {-# INLINE [1] intersectDim #-}
    {-# INLINE [1] addDim #-}
    {-# INLINE [1] size #-}
    {-# INLINE [1] sizeIsValid #-}
    {-# INLINE [1] toIndex #-}
    {-# INLINE [1] fromIndex #-}
    {-# INLINE [1] inShapeRange #-}
    {-# NOINLINE listOfShape #-}
    {-# NOINLINE shapeOfList #-}
    {-# INLINE     deepSeq #-}

instance ExtShape HPIo where
    subDim (HPIo v1 h1 l1) (HPIo v2 h2 l2) = HPIo (v1 `shiftR` popCount v2) (h1-h2) (l1-l2)
    -- TODO subDim is probably borked for anything non-trivial
    rangeList (HPIo 0 (-1) (-1)) (HPIo v _ _) =
      [ HPIo v' h' l'
      | v' <- sortBy (comparing popCount) [0..v]
      , h' <- [0 .. popCount v -1]
      , l' <- [0 .. popCount v -1]
      ]
      {-
      HPIo 0 0 0 :
      [ HPIo v' h' l'
      | v' <- sortBy (comparing popCount) [0..v]
      , let as = activeBits v'
      , h' <- as
      , l' <- if length as == 1 then as else delete h' as
      ]
      -}
    rangeStream (HPIo 0 (-1) (-1)) (HPIo v _ _) = M.fromList $ reverse
      -- the initial node
      [ HPIo v h' h'
      | h' <- activeBits v
      ] ++
      [ HPIo v' h' l'
      | v' <- sortBy (comparing (Down . popCount)) [0..v]
      , let as = activeBits v'
      , h' <- as
      , l' <- delete h' as
      ] ++ [HPIo 0 (-1) (-1)]
    {-# INLINE [1] subDim #-}
    {-# INLINE     rangeList #-}
    {-# INLINE     rangeStream #-}



-- * distance scoring

data Distances a = Distances
               { distances :: !(PAZ.Unboxed (Z:.Int:.Int) a)
               , names     :: !(V.Vector String)
               }
               deriving (Show)

genDistances :: forall k . (Num k, Read k, VU.Unbox k) => String -> Distances k
genDistances s = Distances ds ns where
  ds = fromAssocs (Z:.0:.0) (Z:.(k-1):.(k-1)) (-999999)
        [ ((Z:.i:.j) , read z) | (i,jzs) <- xs, (j,z) <- jzs ]
  k  = length xs
  xs :: [ (Int,[(Int,String)]) ]
  xs = zip [0..]        -- add row index
     . map (zip [0..])  -- add column index
     . tail             -- kill the header row
     . map tail $ ss    -- kill the header column
  ns = V.fromList . tail . map head $ ss
  ss = map words . lines $ s

-- * grammar

data Hamiltonian m x r e = Hamiltonian
                       { start :: e -> x
                       , edge  :: e -> e -> x -> x
                       , h :: M.Stream m x -> m r
                       }

makeAlgebraProductH ['h] ''Hamiltonian



type Node = Int
type NodeName = String
type Score a = a
type Path a = [(NodeName,Node,Node,Score a)]

aWeights :: Monad m => Distances Int -> Hamiltonian m Int Int Node
aWeights Distances{..} = Hamiltonian
  { start = \ s -> {- traceShow s -} 0
  , edge = \ b i t -> {- traceShow (b,"->",i,distances PA.! (Z:.b:.i),t) $ -} t + (distances PA.! (Z:.b:.i))
  , h = M.foldl' min 999999
  }

aProb :: Monad m => Distances (Log Double) -> Hamiltonian m (Log Double) (Log Double) Node
aProb Distances{..} = Hamiltonian
  { start = \ s -> 1
  , edge = \ b i t -> t * (distances PA.! (Z:.b:.i))
  , h = M.foldl' (+) 0
  }

-- TODO if backtracking freaks out, use scaled ints

maxProb :: Monad m => Distances Int -> Hamiltonian m Int Int Node
maxProb Distances{..} = Hamiltonian
  { start = \ s -> 0
  , edge = \ b i t -> t + (distances PA.! (Z:.b:.i))
  , h = M.foldl' max (-999999999)
  }

aPretty :: (Num k, Monad m, VU.Unbox k) => Distances k -> Hamiltonian m (Path k) (M.Stream m (Path k)) Node
aPretty Distances{..} = Hamiltonian
  { start = \s -> [(names V.! s, s,s,0)]
  , edge  = \ b i t -> (names V.! i, i,b,distances PA.! (Z:.b:.i)):t
  , h = return . id
  }


grammar Hamiltonian{..} t =
    Z:.
    (t, start <<< Start                    |||
        edge  <<< Boundary % Interface % t ... h
    )

setallbits k = foldl setBit 0 [0..(k-1)]

minWeight :: Distances Int -> Int -> ST s (Z:.Unboxed HPI Int)
minWeight ds k = do
    let bs = setallbits k
    !t' <- PA.newWithM zeroDim (HPI bs 0 0) 1000
    let t = mTblS EmptyOk t'
    runFreezeMTbls $ grammar (aWeights ds) t

maxProbCYK :: Distances Int -> Int -> ST s (Z:.Unboxed HPI Int)
maxProbCYK ds k = do
    let bs = setallbits k
    !t' <- PA.newWithM zeroDim (HPI bs 0 0) (-10000000)
    let t = mTblS EmptyOk t'
    runFreezeMTbls $ grammar (maxProb ds) t

forward :: Distances (Log Double) -> Int -> ST s (Z:.Unboxed HPI (Log Double))
forward ds k = do
    let bs = setallbits k
    !t' <- PA.newWithM zeroDim (HPI bs 0 0) 0
    let t = mTblS EmptyOk t'
    runFreezeMTbls $ grammar (aProb ds) t

backward :: Distances (Log Double) -> Int -> ST s (Z:.Unboxed HPIo (Log Double))
backward ds k = do
    let bs = setallbits k
    !t' <- PA.newWithM zeroDim (HPIo bs 0 0) 1000
    let t = mTblS EmptyOk t'
    runFreezeMTbls $ grammar (aProb ds) t

backtrack :: Distances Int -> Int -> (Node,Node) -> Z:.Unboxed HPI Int -> [Path Int]
backtrack ds k (i,j) (Z:.t') = unId . M.toList . unId . g $ HPI (setallbits k) i j where
  t = btTblS EmptyOk t' g
  (Z:.(_,g)) = grammar (aWeights ds <** aPretty ds) t

backtrackP :: Distances Int -> Int -> (Node,Node) -> Z:.Unboxed HPI Int -> [Path Int]
backtrackP ds k (i,j) (Z:.t') = unId . M.toList . unId . g $ HPI (setallbits k) i j where
  t = btTblS EmptyOk t' g
  (Z:.(_,g)) = grammar (maxProb ds <** aPretty ds) t

hamiltonianPath ds k = [ opts ij | (ij,_) <- mma ] where
    -- (filter ((<99) . snd) $ assocs t,"x",minima,"x",mma) where
  (Z:.t) = runST $ minWeight ds k
--  bs = backtrack ds k (Z:.t)
  minima = [ ((i,j),t PA.! HPI (setallbits k) i j) | i <- [0.. (k-1)], j <- delete i [0 .. (k-1)] ]
  mma = let m = minimumBy (comparing snd) minima in filter ((snd m ==) . snd) minima
  opts (i,j) = ( (t PA.! HPI (setallbits k) i j), backtrack ds k (i,j) (Z:.t) )

maxProbPath ds k = [ opts ij | (ij,_) <- mma ] where
    -- (filter ((<99) . snd) $ assocs t,"x",minima,"x",mma) where
  (Z:.t) = runST $ maxProbCYK ds k
  maxima = [ ((i,j),t PA.! HPI (setallbits k) i j) | i <- [0.. (k-1)], j <- delete i [0 .. (k-1)] ]
  mma = let m = maximumBy (comparing snd) maxima in filter ((snd m ==) . snd) maxima
  opts (i,j) = ( (t PA.! HPI (setallbits k) i j), backtrackP ds k (i,j) (Z:.t) )

probPaths ds k = (zedF,zedB, tb) where
  (Z:.tf) = runST $ forward  ds k -- table, forward
  (Z:.tb) = runST $ backward ds k -- table, backward
  zedF = [ tf PA.! HPI  (setallbits k) i j | i <- [0.. (k-1)], j <- delete i [0 .. (k-1)] ]
  zedB = [ tb PA.! HPIo (setallbits k) {-(0 .|. i)-} i i | i <- [0.. (k-1)] ]

probPathsX ds nodes = {- traceShow (tf PA.! HPI 1 0 0,zedF,pkl) -} (exp $ Log.ln $ zedF, exp $ Log.ln $ Log.sum $ map snd pkl, pkl) where
  (Z:.tf) = runST $ forward  ds nodes
  zedF = Log.sum [ tf PA.! HPI  xb i j | i <- [0.. (nodes-1)], j <- [0 .. (nodes-1)] ]
  xb = setallbits nodes
  pkl  = [ ((k,l), Log.sum [ let z  = tf PA.! HPI ik i k * tf PA.! HPI kl k l * tf PA.! HPI lj l j
                                 z' = tf PA.! HPI ki k i * tf PA.! HPI lk l k * tf PA.! HPI jl j l
                             in  z -- traceShow (k,l,i,j,ik,kl,lj,z) $ if [i]==activeBits ik || [j]==activeBits jl then traceShow ('x') 2*z else z
                           | i  <- [0 .. (nodes-1)] \\ [l]
                           , j  <- [0 .. (nodes-1)] \\ [k]
                           , ik <- [ 0 .. xb ]
                           , let ki = ik
                           , i `elem` activeBits ik
                           , k `elem` activeBits ik
                           , let kl = setbits 0 [k,l]
                                 lk = kl
                                 lj = xb `clearbits` activeBits ik
                                 jl = lj
                           , j `elem` activeBits lj
                           , l `elem` activeBits lj
                           ] /zedF)
         | k <- [0 .. (nodes-1)]
         , l <- [0 .. (nodes-1)]
         ]

clearbits, setbits :: Int -> [Int] -> Int
clearbits x bs = foldl' clearBit x bs
setbits   x bs = foldl' setBit   x bs

type instance TblConstraint HPI  = TableConstraint
type instance TblConstraint HPIo = TableConstraint

instance IxStaticVar HPI where
    type IxSV HPI = StaticVariable
    initialSV _   = Static

instance IxStaticVar HPIo where
    type IxSV HPIo = StaticVariable
    initialSV _    = Static

-- * MTbl

instance Element ls HPI => Element (ls :!: MTblMulti m arr x HPI) HPI where
    data Elm (ls :!: MTblMulti m arr x HPI) HPI = ElmMTblHPI !x !HPI !(Elm ls HPI)
    type Arg (ls :!: MTblMulti m arr x HPI)     = Arg ls :. x
    getArg (ElmMTblHPI x _ ls) = getArg ls :. x
    getIdx (ElmMTblHPI _ i _ ) = i
    {-# INLINE getArg #-}
    {-# INLINE getIdx #-}

instance Element ls HPIo => Element (ls :!: MTblMulti m arr x HPIo) HPIo where
    data Elm (ls :!: MTblMulti m arr x HPIo) HPIo = ElmMTblHPIo !x !HPIo !(Elm ls HPIo)
    type Arg (ls :!: MTblMulti m arr x HPIo)      = Arg ls :. x
    getArg (ElmMTblHPIo x _ ls) = getArg ls :. x
    getIdx (ElmMTblHPIo _ i _ ) = i
    {-# INLINE getArg #-}
    {-# INLINE getIdx #-}

instance
    ( Monad m
    , PrimMonad m
    , Element ls HPI
    , MkStream m ls HPI
    , MPrimArrayOps arr HPI x
    , Show x
    ) => MkStream m (ls :!: MTblMulti m arr x HPI) HPI where
    mkStream (ls :!: MTbl c t) Static hpi
      = M.mapM (\s -> let i = getIdx s
                      in  PA.readM t i >>= \z -> {- traceShow ("mkStream/MTbl",hpi,i,z) $ -} return $ ElmMTblHPI z i s)
      $ mkStream ls (Variable Check Nothing) hpi

instance
    ( Monad m
    , PrimMonad m
    , Element ls HPIo
    , MkStream m ls HPIo
    , MPrimArrayOps arr HPIo x
    , Show x
    ) => MkStream m (ls :!: MTblMulti m arr x HPIo) HPIo where
    mkStream (ls :!: MTbl c t) Static hpi
      = M.mapM (\s -> let i = getIdx s
                      in  PA.readM t i >>= \z -> traceShow ("mkStream/MTbl",hpi,i,z) $ return $ ElmMTblHPIo z i s)
      $ mkStream ls (Variable Check Nothing) hpi

-- * BtTbl

instance Element ls HPI => Element (ls :!: BtTblMulti m arr x HPI b) HPI where
    data Elm (ls :!: BtTblMulti m arr x HPI b) HPI = ElmBtTblHPI !x !(m (M.Stream m b)) !HPI !(Elm ls HPI)
    type Arg (ls :!: BtTblMulti m arr x HPI b)     = Arg ls :. (x, m (M.Stream m b))
    getArg (ElmBtTblHPI x s _ ls) = getArg ls :. (x,s)
    getIdx (ElmBtTblHPI _ _ i _ ) = i
    {-# INLINE getArg #-}
    {-# INLINE getIdx #-}

instance
    ( Monad m
    , Element ls HPI
    , MkStream m ls HPI
    , PrimArrayOps arr HPI x
    ) => MkStream m (ls :!: BtTblMulti m arr x HPI b) HPI where
      mkStream (ls :!: BtTbl c t f) Static hpi
        = M.map (\s -> let i = getIdx s
                           d = t PA.! i
                       in ElmBtTblHPI d (f i) i s)
        $ mkStream ls (Variable Check Nothing) hpi
      {-# INLINE mkStream #-}

-- * Boundary

-- | The boundary. The new node added to be visited. If @X_ks -> Y@ has @k
-- \in ks@ then afterwards @k@ is not in @ks@ anymore.

data Boundary = Boundary

instance Build Boundary

instance (Element ls HPI) => Element (ls :!: Boundary) HPI where
    data Elm (ls :!: Boundary) HPI = ElmBoundary !Int !HPI !(Elm ls HPI)
    type Arg (ls :!: Boundary)     = Arg ls :. Int
    getArg (ElmBoundary x _ ls) = getArg ls :. x
    getIdx (ElmBoundary _ i _ ) = i
    {-# INLINE getArg #-}
    {-# INLINE getIdx #-}

instance (Element ls HPIo) => Element (ls :!: Boundary) HPIo where
    data Elm (ls :!: Boundary) HPIo = ElmBoundaryO !Int !HPIo !(Elm ls HPIo)
    type Arg (ls :!: Boundary)      = Arg ls :. Int
    getArg (ElmBoundaryO x _ ls) = getArg ls :. x
    getIdx (ElmBoundaryO _ i _ ) = i
    {-# INLINE getArg #-}
    {-# INLINE getIdx #-}

instance
    ( Monad m
    , Element ls HPI
    , MkStream m ls HPI
    ) => MkStream m (ls :!: Boundary) HPI where
      mkStream (ls :!: Boundary) sv ix = M.flatten mk step Unknown $ mkStream ls sv ix where
        mk s = let i  = getIdx s
                   vs = delete (hpiLast i) $ delete (hpiHead i) $ activeBits $ hpiVis i
                   as = if length (activeBits $ hpiVis i) == 2 then [hpiHead i] else vs
               in  return (s :. hpiVis i `clearBit` hpiLast i `setBit` hpiHead i :. hpiHead i :. as)
        step (_ :. _ :. _ :. [])     = return $ M.Done
        step (s :. v :. h :. (k:ks)) = {- traceShow ("mkStream/Boundary",v,h,k) $ -} return $ M.Yield (ElmBoundary k (HPI v h k) s) (s:.v:.h:.ks)
        {-# INLINE [1] mk #-}
        {-# INLINE [1] step #-}
      {-# INLINE mkStream #-}

instance
    ( Monad m
    , Element ls HPIo
    , MkStream m ls HPIo
    ) => MkStream m (ls :!: Boundary) HPIo where
      mkStream (ls :!: Boundary) sv ix = M.flatten mk step Unknown $ mkStream ls sv ix where
        mk s = let i  = getIdx s
                   vs = delete (hpioLast i) $ delete (hpioHead i) $ activeBits $ hpioVis i
                   as = if length (activeBits $ hpioVis i) == 2 then [hpioHead i] else vs
               in  return (s :. hpioVis i `clearBit` hpioLast i `setBit` hpioHead i :. hpioHead i :. as)
        step (_ :. _ :. _ :. [])     = return $ M.Done
        step (s :. v :. h :. (k:ks)) = {- traceShow ("mkStream/Boundary",v,h,k) $ -} return $ M.Yield (ElmBoundaryO k (HPIo v h k) s) (s:.v:.h:.ks)
        {-# INLINE [1] mk #-}
        {-# INLINE [1] step #-}
      {-# INLINE mkStream #-}

-- | The interface. For a production @X->Y@ this is the node visited last
-- in @X@!

data Interface = Interface

instance Build Interface

instance (Element ls HPI) => Element (ls :!: Interface) HPI where
    data Elm (ls :!: Interface) HPI = ElmInterface !Int !HPI !(Elm ls HPI)
    type Arg (ls :!: Interface)     = Arg ls :. Int
    getArg (ElmInterface x _ ls) = getArg ls :. x
    getIdx (ElmInterface _ i _ ) = i
    {-# INLINE getArg #-}
    {-# INLINE getIdx #-}

instance (Element ls HPIo) => Element (ls :!: Interface) HPIo where
    data Elm (ls :!: Interface) HPIo = ElmInterfaceO !Int !HPIo !(Elm ls HPIo)
    type Arg (ls :!: Interface)      = Arg ls :. Int
    getArg (ElmInterfaceO x _ ls) = getArg ls :. x
    getIdx (ElmInterfaceO _ i _ ) = i
    {-# INLINE getArg #-}
    {-# INLINE getIdx #-}

instance
    ( Monad m
    , Element ls HPI
    , MkStream m ls HPI
    ) => MkStream m (ls :!: Interface) HPI where
      mkStream (ls :!: Interface) sv ix@(HPI v h l) = M.map f $ mkStream ls sv ix where
        f s = ElmInterface l (getIdx s) s
        {-# INLINE f #-}
      {-# INLINE mkStream #-}

instance
    ( Monad m
    , Element ls HPIo
    , MkStream m ls HPIo
    ) => MkStream m (ls :!: Interface) HPIo where
      mkStream (ls :!: Interface) sv ix@(HPIo v h l) = M.map f $ mkStream ls sv ix where
        f s = ElmInterfaceO l (getIdx s) s
        {-# INLINE f #-}
      {-# INLINE mkStream #-}



-- * Starting node
--
-- TODO weird mkStream instance

data Start = Start

instance Build Start

instance Element ls HPI => Element (ls :!: Start) HPI where
    data Elm (ls :!: Start) HPI = ElmStartHPI !Int !HPI !(Elm ls HPI)
    type Arg (ls :!: Start)     = Arg ls :. Int
    getArg (ElmStartHPI x _ l) = getArg l :. x
    getIdx (ElmStartHPI _ i _) = i
    {-# INLINE getArg #-}
    {-# INLINE getIdx #-}

instance Element ls HPIo => Element (ls :!: Start) HPIo where
    data Elm (ls :!: Start) HPIo = ElmStartHPIo !Int !HPIo !(Elm ls HPIo)
    type Arg (ls :!: Start)      = Arg ls :. Int
    getArg (ElmStartHPIo x _ l) = getArg l :. x
    getIdx (ElmStartHPIo _ i _) = i
    {-# INLINE getArg #-}
    {-# INLINE getIdx #-}

instance (Monad m, MkStream m ls HPI) => MkStream m (ls :!: Start) HPI where
    mkStream (ls :!: Start) Static hpi@(HPI v h l)
      = M.map (ElmStartHPI h hpi)
      $ M.filter (const (h==l && v>0))
      $ mkStream ls Static hpi
    {-# INLINE mkStream #-}

instance (Monad m, MkStream m ls HPIo) => MkStream m (ls :!: Start) HPIo where
    mkStream (ls :!: Start) Static hpi@(HPIo v h l)
      = M.map (ElmStartHPIo h hpi)
      $ M.filter (const (h==l && v>0))
      $ mkStream ls Static hpi
    {-# INLINE mkStream #-}

-- * S

instance (Monad m) => MkStream m S HPI where
    mkStream S _ hpi = M.singleton $ ElmS hpi
    {-# INLINE mkStream #-}

instance (Monad m) => MkStream m S HPIo where
    mkStream S _ hpi = M.singleton $ ElmS hpi
    {-# INLINE mkStream #-}

