{-# LANGUAGE RecordWildCards #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE TypeSynonymInstances #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE MultiParamTypeClasses #-}

-- |
--
-- TODO allow @nil :: Int -> x@ in makeAlgebraProductH

module ADP.Fusion.Examples.TSP where

import qualified Data.Vector.Fusion.Stream.Monadic as M
import           Data.Array.Repa.Index
import           Control.Monad.ST
import           Data.Strict.Tuple
import           Data.Vector.Fusion.Stream.Size
import           Data.List (delete)
import           Data.Bits (clearBit,popCount)

import           ADP.Fusion
import           Data.PrimitiveArray as PA

import           Data.Array.Repa.Index.BitSet
--import           ADP.Fusion.Permutation




data TSP m x r e k p = TSP
               { nil       :: e -> x
               , edge      :: k -> p -> x -> x
--               , finalEdge :: p -> x -> x
               , h         :: M.Stream m x -> m r
               }

makeAlgebraProductH ['h] ''TSP

type Node = Int
type Edge = (Node,Node)

addEdge :: Monad m => TSP m Int Int () Node Node
addEdge = TSP
  { nil  = \ () -> 0
  , edge = \ t s f -> s + cost f t    -- to from score
--  , finalEdge = \ f s -> s + cost f 0
  , h    = M.foldl' min 999999
  } where cost f t = abs $ f - t
{-# INLINE addEdge #-}

pretty :: Monad m => TSP m [(Int,Int)] (M.Stream m [(Int,Int)]) () Node Node
pretty = TSP
  { nil  = \ ()  -> []
  , edge = \ t f s -> (t,f):s
--  , finalEdge = \ f s -> 0:f:s
  , h    = return . id
  }
{-# INLINE pretty #-}

grammar TSP{..} t =
    Z:.
    (t, nil  <<< Empty                |||       -- the score of being at city '0'
        edge <<< Boundary % t % Final ... h     -- unset the city bit we are at, get the last city visited previously, and the optimal score and at said last city
    )
    {-
    :.
    (s, finalEdge <<< t ... h                   -- move back to city '0'
    )
    -}
{-# INLINE grammar #-}

forward :: Int -> ST s (Z:.Unboxed TspIndex Int)
forward k = do
    !t' <- PA.newWithM undefined undefined (-999999)
    let t = mTblS EmptyOk t'
    runFreezeMTbls $ grammar addEdge t
{-# NOINLINE forward #-}





type instance TblConstraint TspIndex = TableConstraint

instance IxStaticVar TspIndex where
    type IxSV TspIndex = StaticVariable

instance Element ls TspIndex => Element (ls :!: MTblMulti m arr x TspIndex) TspIndex where
    type Arg (ls :!: MTblMulti m arr x TspIndex) = Arg ls :. x

instance (Monad m, Element ls TspIndex, MkStream m ls TspIndex) => MkStream m (ls :!: MTblMulti m arr x TspIndex) TspIndex where



instance Element ls TspIndex => Element (ls :!: Empty) TspIndex where
    data Elm (ls :!: Empty) TspIndex = ElmEmptyTI !TspIndex (Elm ls TspIndex)
    type Arg (ls :!: Empty)          = Arg ls :. ()
    getArg (ElmEmptyTI _ ls) = getArg ls :. ()
    getIdx (ElmEmptyTI i _ ) = i
    {-# INLINE getArg #-}
    {-# INLINE getIdx #-}

instance (Monad m) => MkStream m S TspIndex where
    mkStream S _ i = M.singleton (ElmS i)
    -- TODO variable / static ? (actually yes!)
    {-# INLINE mkStream #-}

instance (Monad m, MkStream m ls TspIndex) => MkStream m (ls :!: Empty) TspIndex where
    mkStream (ls :!: Empty) Static ix
      = M.map (ElmEmptyTI ix)
      $ mkStream ls Static ix
    mkStream _ _ _ = error "mkStream Empty/TspIndex called with Variable"
    {-# INLINE mkStream #-}



-- * Remove a boundary element.

-- |

data Boundary = Boundary

instance Build Boundary

instance (Element ls TspIndex) => Element (ls :!: Boundary) TspIndex where
    data Elm (ls :!: Boundary) TspIndex = ElmBoundary !Int !TspIndex !(Elm ls TspIndex)
    type Arg (ls :!: Boundary)          = Arg ls :. Int
    getArg (ElmBoundary x _ ls) = getArg ls :. x
    getIdx (ElmBoundary _ i _ ) = i
    {-# INLINE getArg #-}
    {-# INLINE getIdx #-}

instance (Monad m, Element ls TspIndex, MkStream m ls TspIndex) => MkStream m (ls :!: Boundary) TspIndex where
    mkStream (ls :!: Boundary) sv ix = M.flatten mk step Unknown $ mkStream ls sv ix where
      mk s = let i = getIdx s in return (s :. (final i) `clearBit` (visited i) :. delete (final i) (activeBits $ visited i))
      step (_:._:.[]) = return $ M.Done
      step (s:.i:.(k:ks)) = return $ M.Yield (ElmBoundary k (TI i k) s) (s:.i:.ks)
      -- TODO (k:ks) to be replaced by a bitset
      {-# INLINE [1] mk #-}
      {-# INLINE [1] step #-}



-- ** Extract the final element of a TSP index with dedicated end. Will
-- /not/ remove said element, just "show" it -- as this is the 'final' part
-- of the index we hand through.

-- | extract the current final element

data Final = Final

instance (Element ls TspIndex) => Element (ls :!: Final) TspIndex where
    data Elm (ls :!: Final) TspIndex = ElmFinal !Int !TspIndex !(Elm ls TspIndex)
    type Arg (ls :!: Final)          = Arg ls :. Int
    getArg (ElmFinal x _ ls) = getArg ls :. x
    getIdx (ElmFinal _ i _ ) = i
    {-# INLINE getArg #-}
    {-# INLINE getIdx #-}

instance (Monad m, Element ls TspIndex, MkStream m ls TspIndex) => MkStream m (ls :!: Final) TspIndex where
    mkStream (ls :!: Final) sv ix@(TI _ final) = M.map (ElmFinal final ix) $ mkStream ls sv ix
    {-# INLINE mkStream #-}

