{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}

-- | 

module Data.Array.Repa.Index.BitSet where

import           Data.Array.Repa.Shape
import           Data.Array.Repa.Index
import qualified Data.Vector.Fusion.Stream.Monadic as M
import           Data.List (sortBy)
import           Data.Bits
import           Data.Ord (comparing)

import           Data.Array.Repa.ExtShape



-- * full TSP index

-- |
--
-- TODO we want make this sparse by not enconding @final@ the index, but
-- rather the numbor of the @final@ visited bit. Instead of a multiplier of
-- the maximum of @final@ in each step, we need a lot less space if only
-- few bits are set -- at the cost of significant lookup time. (Need
-- efficient bit operations here!)
--
-- TODO 'rangeStream' needs an efficient implementation

data TspIndex = TI { visited :: !Int, final :: !Int }
              deriving (Show,Eq)

instance Shape TspIndex where
    rank _ = 1
    zeroDim = TI 0 (-1)
    unitDim = TI 1 0
    intersectDim = error "intersectDim"
    addDim = error "addDim"
    size (TI v _) = v * popCount v
    sizeIsValid (TI v _) = v < (maxBound `div` popCount v)
    toIndex (TI m _) (TI v f) = m * f + v
    fromIndex = error "fromIndex"
    inShapeRange = error "inShapeRange"
    listOfShape = error "listOfShape"
    shapeOfList = error "shapeOfList"
    deepSeq (TI v f) n = v `seq` f `seq` n
    {-# INLINE rank #-}
    {-# INLINE zeroDim #-}
    {-# INLINE unitDim #-}
    {-# INLINE intersectDim #-}
    {-# INLINE addDim #-}
    {-# INLINE size #-}
    {-# INLINE sizeIsValid #-}
    {-# INLINE toIndex #-}
    {-# INLINE fromIndex #-}
    {-# INLINE inShapeRange #-}
    {-# INLINE listOfShape #-}
    {-# INLINE shapeOfList #-}
    {-# INLINE deepSeq #-}

instance ExtShape TspIndex where
    subDim = error "subDim"
    rangeList = error "rangeList"
    rangeStream (TI 0 (-1)) (TI t _) = M.fromList [ TI v f
                                                  | v <- sortBy (comparing popCount) [0..t]
                                                  , f <- activeBits v
                                                  ]
    {-# INLINE subDim #-}
    {-# INLINE rangeList #-}
    {-# INLINE rangeStream #-}



-- * Bitsets

-- | The 'rangeStream' function is currently implemented in a rather ...
-- moronic way.

newtype BitSet = BitSet { bits :: Int }
                 deriving (Eq,Ord,Bounded,Show,FiniteBits,Bits)

activeBits :: Int -> [Int]
activeBits i = [ k | k<-[0..finiteBitSize i -1], testBit i k]
{-# INLINE activeBits #-}

instance Shape BitSet where
    rank _ = 1
    zeroDim = BitSet 0
    unitDim = BitSet 1
    intersectDim = error "intersectDim"
    addDim = error "addDim"
    size (BitSet v) = v
    sizeIsValid = (< maxBound)
    toIndex (BitSet _max) (BitSet v) = v
    fromIndex = error "fromIndex"
    inShapeRange = error "inShapeRange"
    listOfShape = error "listOfShape"
    shapeOfList = error "shapeOfList"
    deepSeq (BitSet v) n = v `seq` n
    {-# INLINE rank #-}
    {-# INLINE zeroDim #-}
    {-# INLINE unitDim #-}
    {-# INLINE intersectDim #-}
    {-# INLINE addDim #-}
    {-# INLINE size #-}
    {-# INLINE sizeIsValid #-}
    {-# INLINE toIndex #-}
    {-# INLINE fromIndex #-}
    {-# INLINE inShapeRange #-}
    {-# INLINE listOfShape #-}
    {-# INLINE shapeOfList #-}
    {-# INLINE deepSeq #-}

instance ExtShape BitSet where
    subDim = error "subDim"
    rangeList = error "rangeList"
    rangeStream (BitSet 0) (BitSet t) = M.map BitSet . M.fromList . sortBy (comparing popCount) $ [0 .. t]
    {-# INLINE subDim #-}
    {-# INLINE rangeList #-}
    {-# INLINE rangeStream #-}

{-
instance (Shape sh) => Shape (sh:.BitSet) where
-}

