{-# LANGUAGE DeriveDataTypeable #-}
{-# LANGUAGE RecordWildCards #-}
{-# LANGUAGE ScopedTypeVariables #-}

module Main where

import Control.Monad
import Text.Printf
import qualified Data.Vector as V
import Data.List (sortBy,genericLength)
import Data.Ord (comparing)
import GHC.Exts (Down(..))
import Data.List.Split
import qualified Data.Map as M
import System.Console.CmdArgs
import Numeric.Log as Log

import qualified Data.PrimitiveArray as PA

import qualified ADP.Fusion.Examples.HamiltonianPath as HP



data Options
    = MinWeight
    { distances :: FilePath
    , maxnodes  :: Maybe Int
    }
    | MaxProb
    { distances :: FilePath
    , maxnodes  :: Maybe Int
    }
    | Prob
    { distances :: FilePath
    , maxnodes  :: Maybe Int
    , scale     :: Double
    , rowfix    :: Bool
    , scaleapd  :: Bool
    }
    deriving (Show,Data,Typeable)

minweight = MinWeight
  { distances = "distance.mat" &= typFile &= help "which distance matrix to use"
  , maxnodes  = Nothing &= help "debug option: restrict number of nodes to use"
  }

maxprob = MaxProb
  {
  }

prob = Prob
  { scale = 1 &= help "scaling factor (or pseudo-temperature)"
  , rowfix   = False &= help "don't standardize the rows"
  , scaleapd = False &= help "scale by average pairwise distance multiplied by the scale parameter"
  }

main = do
    o <- cmdArgs $ modes [minweight,maxprob,prob &= auto] &= verbosity
    case o of
      MinWeight{..} -> do
        d <- HP.genDistances `fmap` (readFile distances)
        let x = maybe (V.length $ HP.names d) id $ maxnodes
        let m = HP.hamiltonianPath d x
        forM_ m $ \(s,ps) -> do
          forM_ ps $ \p -> do
            forM_ p $ \(name,myIndex,nextIndex,pathweight) -> do
              printf "%5s   %3d   " name pathweight
            putStrLn ""
          putStrLn ""
        return ()
      MaxProb{..} -> do
        d :: HP.Distances Double <- HP.genDistances `fmap` (readFile distances)
        let x = maybe (V.length $ HP.names d) id $ maxnodes
            d' :: HP.Distances Int = d { HP.distances = PA.map (round . (*scl) . log . max (1 / scl)) $ HP.distances d }
            m = HP.maxProbPath d' x
            scl = 1000 :: Double
        print d'
        forM_ m $ \(s,ps) -> do
          forM_ ps $ \p -> do
            forM_ p $ \(name,myIndex,nextIndex,pathweight) -> do
              printf "%5s   %5.2f   " name ((exp . (/scl) . fromIntegral $ pathweight) :: Double)
            putStrLn ""
          putStrLn ""
        return ()
      Prob{..} -> do
        d :: HP.Distances Int <- HP.genDistances `fmap` (readFile distances)
        let x = maybe (V.length $ HP.names d) id $ maxnodes
            s = if scaleapd then let ds = map fromIntegral . PA.toList $ HP.distances d in scale * Prelude.sum ds / genericLength ds
                            else scale
            d' :: HP.Distances (Log Double) = d { HP.distances = PA.map (Exp . (/ s) . negate . fromIntegral) $ HP.distances d }
            (_,_,os) = HP.probPathsX d' x
            mos = M.fromList os
        putStr "        "
        forM_ [0 .. x-1] $ \i -> printf "%5s   " (HP.names d V.! i)
        putStrLn ""
        forM_ [0 .. x-1] $ \i -> do
          printf "%5s   " (HP.names d V.! i)
          let isum = (exp . ln $ Log.sum [ mos M.! (i,j) | j <- [0 .. x-1] ])
          forM_ [0 .. x-1] $ \j -> printf "%5.2f   " (let v = (exp . ln $ mos M.! (i,j)) in if rowfix then v / isum else v)
          putStrLn ""
        putStrLn ""

